"""The Test file for the linter class."""

import pytest
import logging

from sqlfluff.core.linter import LintedFile
from sqlfluff.core.parser.markers import PositionMarker
from sqlfluff.core.parser.segments import (
    FixPatch,
    RawSegment,
    BaseSegment,
    TemplateSegment,
)
from sqlfluff.core.parser.segments.raw import SourceFix
from sqlfluff.core.templaters import RawFileSlice, TemplatedFile
from sqlfluff.core.templaters.base import TemplatedFileSlice


@pytest.mark.parametrize(
    "source_slices,source_patches,raw_source_string,expected_result",
    # NOTE: For all of these examples we're not setting the patch_category
    # of the fix patches. They're not used at this step so irrelevant for
    # testing.
    [
        # Trivial example
        ([slice(0, 1)], [], "a", "a"),
        # Simple replacement
        (
            [slice(0, 1), slice(1, 2), slice(2, 3)],
            [FixPatch(slice(1, 2), "d", "", slice(1, 2), "b", "b")],
            "abc",
            "adc",
        ),
        # Simple insertion
        (
            [slice(0, 1), slice(1, 1), slice(1, 2)],
            [FixPatch(slice(1, 1), "b", "", slice(1, 1), "", "")],
            "ac",
            "abc",
        ),
        # Simple deletion
        (
            [slice(0, 1), slice(1, 2), slice(2, 3)],
            [FixPatch(slice(1, 2), "", "", slice(1, 2), "b", "b")],
            "abc",
            "ac",
        ),
        # Illustrative templated example (although practically at
        # this step, the routine shouldn't care if it's templated).
        (
            [slice(0, 2), slice(2, 7), slice(7, 9)],
            [FixPatch(slice(2, 3), "{{ b }}", "", slice(2, 7), "b", "{{b}}")],
            "a {{b}} c",
            "a {{ b }} c",
        ),
    ],
)
def test__linted_file__build_up_fixed_source_string(
    source_slices, source_patches, raw_source_string, expected_result, caplog
):
    """Test _build_up_fixed_source_string.

    This is part of fix_string().
    """
    with caplog.at_level(logging.DEBUG, logger="sqlfluff.linter"):
        result = LintedFile._build_up_fixed_source_string(
            source_slices, source_patches, raw_source_string
        )
    assert result == expected_result


@pytest.mark.parametrize(
    "source_patches,source_only_slices,raw_source_string,expected_result",
    # NOTE: For all of these examples we're not setting the patch_category
    # of the fix patches. They're not used at this step so irrelevant for
    # testing.
    [
        # Trivial example.
        # No edits in a single character file. Slice should be one
        # character long.
        ([], [], "a", [slice(0, 1)]),
        # Simple replacement.
        # We've yielded a patch to change a single character. This means
        # we should get only slices for that character, and for the
        # unchanged file around it.
        (
            [FixPatch(slice(1, 2), "d", "", slice(1, 2), "b", "b")],
            [],
            "abc",
            [slice(0, 1), slice(1, 2), slice(2, 3)],
        ),
        # Templated no fixes.
        # A templated file, but with no fixes, so no subdivision of the
        # file is required and we should just get a single slice.
        (
            [],
            [],
            "a {{ b }} c",
            [slice(0, 11)],
        ),
        # Templated example with a source-only slice.
        # A templated file, but with no fixes, so no subdivision of the
        # file is required and we should just get a single slice. While
        # there is handling for "source only" slices like template
        # comments, in this case no additional slicing is required
        # because no edits have been made.
        (
            [],
            [RawFileSlice("{# b #}", "comment", 2)],
            "a {# b #} c",
            [slice(0, 11)],
        ),
        # Templated fix example with a source-only slice.
        # We're making an edit adjacent to a source only slice. Edits
        # _before_ source only slices currently don't trigger additional
        # slicing. This is fine.
        (
            [FixPatch(slice(0, 1), "a ", "", slice(0, 1), "a", "a")],
            [RawFileSlice("{# b #}", "comment", 1)],
            "a{# b #}c",
            [slice(0, 1), slice(1, 9)],
        ),
        # Templated fix example with a source-only slice.
        # We've made an edit directly _after_ a source only slice
        # which should trigger the logic to ensure that the source
        # only slice isn't included in the source mapping of the
        # edit.
        # TODO: given that the logic is based on the _type_
        # of the slice (e.g. comment), would we handle a
        # template tag which returns an empty string correctly?
        (
            [FixPatch(slice(1, 2), " c", "", slice(8, 9), "c", "c")],
            [RawFileSlice("{# b #}", "comment", 1)],
            "a{# b #}cc",
            [slice(0, 1), slice(1, 8), slice(8, 9), slice(9, 10)],
        ),
        # Templated example with a source-only slice.
        # Here we're making the fix to the templated slice. This
        # checks that we don't duplicate or fumble the slice
        # generation when we're explicitly trying to edit the source.
        # TODO: Should we be using the fix type (e.g. "source")
        # to somehow determine whether the fix is "safe"?
        (
            [FixPatch(slice(2, 2), "{# fixed #}", "", slice(2, 9), "", "")],
            [RawFileSlice("{# b #}", "comment", 2)],
            "a {# b #} c",
            [slice(0, 2), slice(2, 9), slice(9, 11)],
        ),
        # Illustrate potential templating bug (case from L046).
        # In this case we have fixes for all our tempolated sections
        # and they are all close to each other and so may be either
        # skipped or duplicated if the logic is not precise.
        (
            [
                FixPatch(
                    templated_slice=slice(14, 14),
                    fixed_raw="{%+ if true -%}",
                    patch_category="source",
                    source_slice=slice(14, 27),
                    templated_str="",
                    source_str="{%+if true-%}",
                ),
                FixPatch(
                    templated_slice=slice(14, 14),
                    fixed_raw="{{ ref('foo') }}",
                    patch_category="source",
                    source_slice=slice(28, 42),
                    templated_str="",
                    source_str="{{ref('foo')}}",
                ),
                FixPatch(
                    templated_slice=slice(17, 17),
                    fixed_raw="{%- endif %}",
                    patch_category="source",
                    source_slice=slice(43, 53),
                    templated_str="",
                    source_str="{%-endif%}",
                ),
            ],
            [
                RawFileSlice(
                    raw="{%+if true-%}",
                    slice_type="block_start",
                    source_idx=14,
                    block_idx=0,
                ),
                RawFileSlice(
                    raw="{%-endif%}",
                    slice_type="block_end",
                    source_idx=43,
                    block_idx=1,
                ),
            ],
            "SELECT 1 from {%+if true-%} {{ref('foo')}} {%-endif%}",
            [
                slice(0, 14),
                slice(14, 27),
                slice(27, 28),
                slice(28, 42),
                slice(42, 43),
                slice(43, 53),
            ],
        ),
    ],
)
def test__linted_file__slice_source_file_using_patches(
    source_patches, source_only_slices, raw_source_string, expected_result, caplog
):
    """Test _slice_source_file_using_patches.

    This is part of fix_string().
    """
    with caplog.at_level(logging.DEBUG, logger="sqlfluff.linter"):
        result = LintedFile._slice_source_file_using_patches(
            source_patches, source_only_slices, raw_source_string
        )
    assert result == expected_result


templated_file_1 = TemplatedFile.from_string("abc")
templated_file_2 = TemplatedFile(
    "{# blah #}{{ foo }}bc",
    "<testing>",
    "abc",
    [
        TemplatedFileSlice("comment", slice(0, 10), slice(0, 0)),
        TemplatedFileSlice("templated", slice(10, 19), slice(0, 1)),
        TemplatedFileSlice("literal", slice(19, 21), slice(1, 3)),
    ],
    [
        RawFileSlice("{# blah #}", "comment", 0),
        RawFileSlice("{{ foo }}", "templated", 10),
        RawFileSlice("bc", "literal", 19),
    ],
)


@pytest.mark.parametrize(
    "tree,templated_file,expected_result",
    [
        # Trivial example
        (
            RawSegment(
                "abc",
                PositionMarker(slice(0, 3), slice(0, 3), templated_file_1),
                "code",
            ),
            templated_file_1,
            [],
        ),
        # Simple literal edit example
        (
            RawSegment(
                "abz",
                PositionMarker(slice(0, 3), slice(0, 3), templated_file_1),
                "code",
            ),
            templated_file_1,
            [FixPatch(slice(0, 3), "abz", "literal", slice(0, 3), "abc", "abc")],
        ),
        # Nested literal edit example
        (
            BaseSegment(
                [
                    RawSegment(
                        "a",
                        PositionMarker(slice(0, 1), slice(0, 1), templated_file_1),
                        "code",
                    ),
                    RawSegment(
                        "b",
                        PositionMarker(slice(1, 2), slice(1, 2), templated_file_1),
                        "code",
                    ),
                    RawSegment(
                        "z",
                        PositionMarker(slice(2, 3), slice(2, 3), templated_file_1),
                        "code",
                    ),
                ]
            ),
            templated_file_1,
            [FixPatch(slice(0, 3), "abz", "literal", slice(0, 3), "abc", "abc")],
        ),
        # More complicated templating example
        (
            BaseSegment(
                [
                    TemplateSegment(
                        PositionMarker(slice(0, 10), slice(0, 0), templated_file_2),
                        "{# blah #}",
                        "comment",
                    ),
                    RawSegment(
                        "a",
                        PositionMarker(slice(10, 20), slice(0, 1), templated_file_2),
                        "code",
                    ),
                    RawSegment(
                        "b",
                        PositionMarker(slice(19, 20), slice(1, 2), templated_file_2),
                        "code",
                    ),
                    RawSegment(
                        "z",
                        PositionMarker(slice(20, 21), slice(2, 3), templated_file_2),
                        "code",
                    ),
                ]
            ),
            templated_file_2,
            [FixPatch(slice(2, 3), "z", "literal", slice(20, 21), "c", "c")],
        ),
        # Templating example with fixes
        (
            BaseSegment(
                [
                    TemplateSegment(
                        PositionMarker(slice(0, 10), slice(0, 0), templated_file_2),
                        "{# blah #}",
                        "comment",
                        source_fixes=[
                            SourceFix("{# fixed #}", slice(0, 10), slice(0, 0))
                        ],
                    ),
                    RawSegment(
                        "a",
                        PositionMarker(slice(10, 19), slice(0, 1), templated_file_2),
                        "code",
                        source_fixes=[
                            SourceFix("{{ bar }}", slice(10, 19), slice(0, 1))
                        ],
                    ),
                    RawSegment(
                        "b",
                        PositionMarker(slice(19, 20), slice(1, 2), templated_file_2),
                        "code",
                    ),
                    RawSegment(
                        "z",
                        PositionMarker(slice(20, 21), slice(2, 3), templated_file_2),
                        "code",
                    ),
                ]
            ),
            templated_file_2,
            [
                FixPatch(
                    slice(0, 0), "{# fixed #}", "source", slice(0, 10), "", "{# blah #}"
                ),
                FixPatch(
                    slice(0, 1), "{{ bar }}", "source", slice(10, 19), "a", "{{ foo }}"
                ),
                FixPatch(slice(2, 3), "z", "literal", slice(20, 21), "c", "c"),
            ],
        ),
    ],
)
def test__linted_file__generate_source_patches(
    tree, templated_file, expected_result, caplog
):
    """Test _generate_source_patches.

    This is part of fix_string().
    """
    with caplog.at_level(logging.DEBUG, logger="sqlfluff.linter"):
        result = LintedFile._generate_source_patches(tree, templated_file)
    assert result == expected_result



"""Tests for templaters."""

import pytest

from sqlfluff.core.templaters import (
    RawTemplater,
    TemplatedFile,
)

from sqlfluff.core.templaters.base import (
    iter_indices_of_newlines,
    RawFileSlice,
    TemplatedFileSlice,
)


@pytest.mark.parametrize(
    "raw_str,positions",
    [
        ("", []),
        ("foo", []),
        ("foo\nbar", [3]),
        ("\nfoo\n\nbar\nfoo\n\nbar\n", [0, 4, 5, 9, 13, 14, 18]),
    ],
)
def test__indices_of_newlines(raw_str, positions):
    """Test iter_indices_of_newlines."""
    assert list(iter_indices_of_newlines(raw_str)) == positions


def test__templater_raw():
    """Test the raw templater."""
    t = RawTemplater()
    instr = "SELECT * FROM {{blah}}"
    outstr, _ = t.process(in_str=instr, fname="test")
    assert instr == str(outstr)


SIMPLE_SOURCE_STR = "01234\n6789{{foo}}fo\nbarss"
SIMPLE_TEMPLATED_STR = "01234\n6789x\nfo\nbarfss"
SIMPLE_SLICED_FILE = [
    TemplatedFileSlice(*args)
    for args in [
        ("literal", slice(0, 10, None), slice(0, 10, None)),
        ("templated", slice(10, 17, None), slice(10, 12, None)),
        ("literal", slice(17, 25, None), slice(12, 20, None)),
    ]
]
SIMPLE_RAW_SLICED_FILE = [
    RawFileSlice(*args)
    for args in [
        ("x" * 10, "literal", 0),
        ("x" * 7, "templated", 10),
        ("x" * 8, "literal", 17),
    ]
]

COMPLEX_SLICED_FILE = [
    TemplatedFileSlice(*args)
    for args in [
        ("literal", slice(0, 13, None), slice(0, 13, None)),
        ("comment", slice(13, 29, None), slice(13, 13, None)),
        ("literal", slice(29, 44, None), slice(13, 28, None)),
        ("block_start", slice(44, 68, None), slice(28, 28, None)),
        ("literal", slice(68, 81, None), slice(28, 41, None)),
        ("templated", slice(81, 86, None), slice(41, 42, None)),
        ("literal", slice(86, 110, None), slice(42, 66, None)),
        ("templated", slice(68, 86, None), slice(66, 76, None)),
        ("literal", slice(68, 81, None), slice(76, 89, None)),
        ("templated", slice(81, 86, None), slice(89, 90, None)),
        ("literal", slice(86, 110, None), slice(90, 114, None)),  #
        ("templated", slice(68, 86, None), slice(114, 125, None)),
        ("literal", slice(68, 81, None), slice(125, 138, None)),  #
        ("templated", slice(81, 86, None), slice(138, 139, None)),
        ("literal", slice(86, 110, None), slice(139, 163, None)),
        ("templated", slice(110, 123, None), slice(163, 166, None)),
        ("literal", slice(123, 132, None), slice(166, 175, None)),
        ("block_end", slice(132, 144, None), slice(175, 175, None)),
        ("literal", slice(144, 155, None), slice(175, 186, None)),
        ("block_start", slice(155, 179, None), slice(186, 186, None)),
        ("literal", slice(179, 189, None), slice(186, 196, None)),
        ("templated", slice(189, 194, None), slice(196, 197, None)),
        ("literal", slice(194, 203, None), slice(197, 206, None)),
        ("literal", slice(179, 189, None), slice(206, 216, None)),
        ("templated", slice(189, 194, None), slice(216, 217, None)),
        ("literal", slice(194, 203, None), slice(217, 226, None)),
        ("literal", slice(179, 189, None), slice(226, 236, None)),
        ("templated", slice(189, 194, None), slice(236, 237, None)),
        ("literal", slice(194, 203, None), slice(237, 246, None)),
        ("block_end", slice(203, 215, None), slice(246, 246, None)),
        ("literal", slice(215, 230, None), slice(246, 261, None)),
    ]
]
COMPLEX_RAW_SLICED_FILE = [
    RawFileSlice(*args)
    for args in [
        # All contain dummy strings for now.
        ("x" * 13, "literal", 0),
        ("x" * 16, "comment", 13),
        ("x" * 15, "literal", 29),
        ("x" * 24, "block_start", 44),
        ("x" * 13, "literal", 68),
        ("x" * 5, "templated", 81),
        ("x" * 24, "literal", 86),
        ("x" * 13, "templated", 110),
        ("x" * 9, "literal", 123),
        ("x" * 12, "block_end", 132),
        ("x" * 11, "literal", 144),
        ("x" * 24, "block_start", 155),
        ("x" * 10, "literal", 179),
        ("x" * 5, "templated", 189),
        ("x" * 9, "literal", 194),
        ("x" * 12, "block_end", 203),
        ("x" * 15, "literal", 215),
    ]
]


@pytest.mark.parametrize(
    "source_str,templated_str,file_slices,in_charpos,out_line_no,out_line_pos",
    [
        # Simple examples
        (SIMPLE_SOURCE_STR, SIMPLE_TEMPLATED_STR, SIMPLE_SLICED_FILE, 0, 1, 1),
        (SIMPLE_SOURCE_STR, SIMPLE_TEMPLATED_STR, SIMPLE_SLICED_FILE, 20, 3, 1),
        (SIMPLE_SOURCE_STR, SIMPLE_TEMPLATED_STR, SIMPLE_SLICED_FILE, 24, 3, 5),
    ],
)
def test__templated_file_get_line_pos_of_char_pos(
    source_str, templated_str, file_slices, in_charpos, out_line_no, out_line_pos
):
    """Test TemplatedFile.get_line_pos_of_char_pos."""
    file = TemplatedFile(
        source_str=source_str,
        templated_str=templated_str,
        sliced_file=file_slices,
        fname="test",
        check_consistency=False,
    )
    res_line_no, res_line_pos = file.get_line_pos_of_char_pos(in_charpos)
    assert res_line_no == out_line_no
    assert res_line_pos == out_line_pos


@pytest.mark.parametrize(
    "templated_position,inclusive,file_slices,sliced_idx_start,sliced_idx_stop",
    [
        (100, True, COMPLEX_SLICED_FILE, 10, 11),
        (13, True, COMPLEX_SLICED_FILE, 0, 3),
        (28, True, COMPLEX_SLICED_FILE, 2, 5),
        # Check end slicing.
        (12, True, SIMPLE_SLICED_FILE, 1, 3),
        (20, True, SIMPLE_SLICED_FILE, 2, 3),
        # Check inclusivity
        (13, False, COMPLEX_SLICED_FILE, 0, 1),
    ],
)
def test__templated_file_find_slice_indices_of_templated_pos(
    templated_position, inclusive, file_slices, sliced_idx_start, sliced_idx_stop
):
    """Test TemplatedFile._find_slice_indices_of_templated_pos."""
    file = TemplatedFile(
        source_str="Dummy String", sliced_file=file_slices, fname="test"
    )
    res_start, res_stop = file._find_slice_indices_of_templated_pos(
        templated_position, inclusive=inclusive
    )
    assert res_start == sliced_idx_start
    assert res_stop == sliced_idx_stop


@pytest.mark.parametrize(
    "in_slice,out_slice,is_literal,file_slices,raw_slices",
    [
        # Simple example
        (
            slice(5, 10),
            slice(5, 10),
            True,
            [TemplatedFileSlice("literal", slice(0, 20, None), slice(0, 20, None))],
            [RawFileSlice("x" * 20, "literal", 0)],
        ),
        # Trimming the end of a literal (with things that follow).
        (
            slice(10, 13),
            slice(10, 13),
            True,
            COMPLEX_SLICED_FILE,
            COMPLEX_RAW_SLICED_FILE,
        ),
        # Unrealistic, but should still work
        (
            slice(5, 10),
            slice(55, 60),
            True,
            [TemplatedFileSlice("literal", slice(50, 70, None), slice(0, 20, None))],
            [RawFileSlice("x" * 50, "literal", 0), ("x" * 20, "literal", 50)],
        ),
        # Spanning a template
        (
            slice(5, 15),
            slice(5, 20),
            False,
            SIMPLE_SLICED_FILE,
            SIMPLE_RAW_SLICED_FILE,
        ),
        # Handling templated
        (
            slice(5, 15),
            slice(0, 25),
            False,
            # NB: Same as SIMPLE_SLICED_FILE, but with different slice types.
            [
                TemplatedFileSlice("templated", slc.source_slice, slc.templated_slice)
                for slc in SIMPLE_SLICED_FILE
            ],
            [
                RawFileSlice(slc.raw, "templated", slc.source_idx)
                for slc in SIMPLE_RAW_SLICED_FILE
            ],
        ),
        # Handling single length slices
        (
            slice(10, 10),
            slice(10, 10),
            True,
            SIMPLE_SLICED_FILE,
            SIMPLE_RAW_SLICED_FILE,
        ),
        (
            slice(12, 12),
            slice(17, 17),
            True,
            SIMPLE_SLICED_FILE,
            SIMPLE_RAW_SLICED_FILE,
        ),
        # Dealing with single length elements
        (
            slice(20, 20),
            slice(25, 25),
            True,
            SIMPLE_SLICED_FILE
            + [TemplatedFileSlice("comment", slice(25, 35, None), slice(20, 20, None))],
            SIMPLE_RAW_SLICED_FILE + [RawFileSlice("x" * 10, "comment", 25)],
        ),
        # Just more test coverage
        (
            slice(43, 43),
            slice(87, 87),
            True,
            COMPLEX_SLICED_FILE,
            COMPLEX_RAW_SLICED_FILE,
        ),
        (
            slice(13, 13),
            slice(13, 13),
            True,
            COMPLEX_SLICED_FILE,
            COMPLEX_RAW_SLICED_FILE,
        ),
        (
            slice(186, 186),
            slice(155, 155),
            True,
            COMPLEX_SLICED_FILE,
            COMPLEX_RAW_SLICED_FILE,
        ),
        # Backward slicing.
        (
            slice(100, 130),
            # NB This actually would reference the wrong way around if we
            # just take the points. Here we should handle it gracefully.
            slice(68, 110),
            False,
            COMPLEX_SLICED_FILE,
            COMPLEX_RAW_SLICED_FILE,
        ),
    ],
)
def test__templated_file_templated_slice_to_source_slice(
    in_slice, out_slice, is_literal, file_slices, raw_slices
):
    """Test TemplatedFile.templated_slice_to_source_slice."""
    file = TemplatedFile(
        source_str="Dummy String",
        sliced_file=file_slices,
        raw_sliced=[
            rs if isinstance(rs, RawFileSlice) else RawFileSlice(*rs)
            for rs in raw_slices
        ],
        fname="test",
        check_consistency=False,
    )
    source_slice = file.templated_slice_to_source_slice(in_slice)
    literal_test = file.is_source_slice_literal(source_slice)
    assert (is_literal, source_slice) == (literal_test, out_slice)


@pytest.mark.parametrize(
    "file,expected_result",
    [
        # Comment example
        (
            TemplatedFile(
                source_str="a" * 20,
                fname="test",
                raw_sliced=[
                    RawFileSlice("a" * 10, "literal", 0),
                    RawFileSlice("{# b #}", "comment", 10),
                    RawFileSlice("a" * 10, "literal", 17),
                ],
                check_consistency=False,
            ),
            [RawFileSlice("{# b #}", "comment", 10)],
        ),
        # Template tags aren't source only.
        (
            TemplatedFile(
                source_str="aaabbbaaa",
                fname="test",
                raw_sliced=[
                    RawFileSlice("aaa", "literal", 0),
                    RawFileSlice("{{ b }}", "templated", 3),
                    RawFileSlice("aaa", "literal", 6),
                ],
                check_consistency=False,
            ),
            [],
        ),
    ],
)
def test__templated_file_source_only_slices(file, expected_result):
    """Test TemplatedFile.source_only_slices."""
    assert file.source_only_slices() == expected_result



"""Implements the PositionMarker class.

This class is a construct to keep track of positions within a file.
"""

from dataclasses import dataclass
from typing import Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from sqlfluff.core.templaters import TemplatedFile  # pragma: no cover


@dataclass(frozen=True)
class PositionMarker:
    """A reference to a position in a file.

    Things to note:
        - This combines the previous functionality of FilePositionMarker
          and EnrichedFilePositionMarker. Additionally it contains a reference
          to the original templated file.
        - It no longer explicitly stores a line number or line position in the
          source or template. This is extrapolated from the templated file as required.
        - Positions in the source and template are with slices and therefore identify
          ranges.
        - Positions within the fixed file are identified with a line number and line
          position, which identify a point.
        - Arithmetic comparisons are on the location in the fixed file.
    """

    source_slice: slice
    templated_slice: slice
    templated_file: "TemplatedFile"
    # If not set, these will be initialised in the post init.
    working_line_no: int = -1
    working_line_pos: int = -1

    def __post_init__(self):
        # If the working position has not been explicitly set
        # then infer it from the position in the templated file.
        # This is accurate up until the point that any fixes have
        # been applied.
        if self.working_line_no == -1 or self.working_line_pos == -1:
            line_no, line_pos = self.templated_position()
            # Use the base method because we're working with a frozen class
            object.__setattr__(self, "working_line_no", line_no)
            object.__setattr__(self, "working_line_pos", line_pos)

    def __str__(self):
        return self.to_source_string()

    def __gt__(self, other):
        return self.working_loc > other.working_loc  # pragma: no cover TODO?

    def __lt__(self, other):
        return self.working_loc < other.working_loc  # pragma: no cover TODO?

    def __ge__(self, other):
        return self.working_loc >= other.working_loc  # pragma: no cover TODO?

    def __le__(self, other):
        return self.working_loc <= other.working_loc  # pragma: no cover TODO?

    @property
    def working_loc(self) -> Tuple[int, int]:
        """Location tuple for the working position."""
        return self.working_line_no, self.working_line_pos

    def working_loc_after(self, raw: str) -> Tuple[int, int]:
        """Location tuple for the working position."""
        return self.infer_next_position(
            raw,
            self.working_line_no,
            self.working_line_pos,
        )

    @classmethod
    def from_point(
        cls,
        source_point: int,
        templated_point: int,
        templated_file: "TemplatedFile",
        **kwargs,
    ):
        """Convenience method for creating point markers."""
        return cls(
            slice(source_point, source_point),
            slice(templated_point, templated_point),
            templated_file,
            **kwargs,
        )

    @classmethod
    def from_child_markers(cls, *markers):
        """Create a parent marker from it's children."""
        source_slice = slice(
            min(m.source_slice.start for m in markers),
            max(m.source_slice.stop for m in markers),
        )
        templated_slice = slice(
            min(m.templated_slice.start for m in markers),
            max(m.templated_slice.stop for m in markers),
        )
        templated_files = {m.templated_file for m in markers}
        if len(templated_files) != 1:  # pragma: no cover
            raise ValueError("Attempted to make a parent marker from multiple files.")
        templated_file = templated_files.pop()
        return cls(source_slice, templated_slice, templated_file)

    def source_position(self) -> Tuple[int, int]:
        """Return the line and position of this marker in the source."""
        return self.templated_file.get_line_pos_of_char_pos(
            self.source_slice.start, source=True
        )

    def templated_position(self) -> Tuple[int, int]:
        """Return the line and position of this marker in the source."""
        return self.templated_file.get_line_pos_of_char_pos(
            self.templated_slice.start, source=False
        )

    @property
    def line_no(self) -> int:
        """Return the line number in the source."""
        return self.source_position()[0]

    @property
    def line_pos(self) -> int:
        """Return the line position in the source."""
        return self.source_position()[1]

    def to_source_string(self) -> str:
        """Make a formatted string of this position."""
        line, pos = self.source_position()
        return f"[L:{line:3d}, P:{pos:3d}]"

    def start_point_marker(self) -> "PositionMarker":
        """Get a point marker from the start."""
        return self.__class__.from_point(
            self.source_slice.start,
            self.templated_slice.start,
            templated_file=self.templated_file,
            # Start points also pass on the working position.
            working_line_no=self.working_line_no,
            working_line_pos=self.working_line_pos,
        )

    def end_point_marker(self) -> "PositionMarker":
        """Get a point marker from the end."""
        return self.__class__.from_point(
            self.source_slice.stop,
            self.templated_slice.stop,
            templated_file=self.templated_file,
        )

    @staticmethod
    def slice_is_point(test_slice):
        """Is this slice a point."""
        return test_slice.start == test_slice.stop

    def is_point(self) -> bool:
        """A marker is a point if it has zero length in templated and source file."""
        return self.slice_is_point(self.source_slice) and self.slice_is_point(
            self.templated_slice
        )

    @staticmethod
    def infer_next_position(raw: str, line_no: int, line_pos: int) -> Tuple[int, int]:
        """Using the raw string provided to infer the position of the next.

        NB: Line position in 1-indexed.
        """
        # No content?
        if not raw:
            return line_no, line_pos
        split = raw.split("\n")
        return (
            line_no + len(split) - 1,
            line_pos + len(raw) if len(split) == 1 else len(split[-1]) + 1,
        )

    def with_working_position(self, line_no: int, line_pos: int) -> "PositionMarker":
        """Copy this position and replace the working position."""
        return self.__class__(
            source_slice=self.source_slice,
            templated_slice=self.templated_slice,
            templated_file=self.templated_file,
            working_line_no=line_no,
            working_line_pos=line_pos,
        )

    def is_literal(self) -> bool:
        """Infer literalness from context.

        is_literal should return True if a fix can be applied across this area
        in the templated file while being confident that the fix is still
        appropriate in the source file. This obviously applies to any slices
        which are the same in the source and the templated files. Slices which
        are zero-length in the source are also "literal" because they can't be
        "broken" by any fixes, because they don't exist in the source. This
        includes meta segments and any segments added during the fixing process.

        This value is used for:
        - Ignoring linting errors in templated sections.
        - Whether `iter_patches` can return without recursing.
        - Whether certain rules (such as L046) are triggered.
        """
        return self.templated_file.is_source_slice_literal(self.source_slice)

    def source_str(self) -> str:
        """Returns the string in the source at this position."""
        return self.templated_file.source_str[self.source_slice]



"""Tests for the depthmap object."""

from sqlfluff.core import Linter

from sqlfluff.utils.reflow.depthmap import DepthMap


def parse_ansi_string(sql, config):
    """Parse an ansi sql string for testing."""
    linter = Linter(config=config)
    return linter.parse_string(sql).tree


def test_reflow_depthmap_from_parent(default_config):
    """Test map construction from a root segment."""
    sql = "SELECT 1"
    root = parse_ansi_string(sql, default_config)

    dm = DepthMap.from_parent(root)

    # We use UUIDS in the depth map so we can't assert their value.
    # What we can do is use them.

    # Check that we get the right depths.
    assert [dm.depth_info[seg.uuid].stack_depth for seg in root.raw_segments] == [
        4,
        4,
        4,
        5,
        3,
        1,
    ]
    # Check they all share the same first three hash and
    # class type elements (except the end of file marker at the end).
    # These should be the file, statement and select statement.
    expected = ({"file", "base"}, {"statement", "base"}, {"select_statement", "base"})
    assert all(
        dm.depth_info[seg.uuid].stack_class_types[:3] == expected
        for seg in root.raw_segments[:-1]
    )
    first_hashes = dm.depth_info[root.raw_segments[0].uuid].stack_hashes[:3]
    assert all(
        dm.depth_info[seg.uuid].stack_hashes[:3] == first_hashes
        for seg in root.raw_segments[:-1]
    )

    # While we're here, test the DepthInfo.common_with method
    select_keyword_di = dm.depth_info[root.raw_segments[0].uuid]
    numeric_one_di = dm.depth_info[root.raw_segments[3].uuid]
    assert len(select_keyword_di.common_with(numeric_one_di)) == 4


def test_reflow_depthmap_from_raws_and_root(default_config):
    """Test that the indirect route is equivalent to the direct route."""
    sql = "SELECT 1"
    root = parse_ansi_string(sql, default_config)

    # Direct route
    dm_direct = DepthMap.from_parent(root)

    # Indirect route.
    dm_indirect = DepthMap.from_raws_and_root(root.raw_segments, root)

    # The depth info dict depends on the sequence so we only need
    # to check those are equal.
    assert dm_direct.depth_info == dm_indirect.depth_info



"""The Test file for The New Parser (Grammar Classes)."""

import pytest
import logging

from sqlfluff.core.parser import (
    KeywordSegment,
    StringParser,
    SymbolSegment,
    RegexParser,
    WhitespaceSegment,
    Indent,
)
from sqlfluff.core.parser.context import RootParseContext
from sqlfluff.core.parser.grammar.anyof import AnySetOf
from sqlfluff.core.parser.segments import EphemeralSegment, BaseSegment
from sqlfluff.core.parser.grammar.base import BaseGrammar
from sqlfluff.core.parser.grammar.noncode import NonCodeMatcher
from sqlfluff.core.parser.grammar import (
    OneOf,
    Sequence,
    GreedyUntil,
    Delimited,
    StartsWith,
    Anything,
    Nothing,
    Ref,
    Conditional,
)
from sqlfluff.core.errors import SQLParseError
from os import getenv

# NB: All of these tests depend somewhat on the KeywordSegment working as planned


def make_result_tuple(result_slice, matcher_keywords, seg_list):
    """Make a comparison tuple for test matching."""
    # No result slice means no match.
    if not result_slice:
        return ()

    return tuple(
        KeywordSegment(elem.raw, pos_marker=elem.pos_marker)
        if elem.raw in matcher_keywords
        else elem
        for elem in seg_list[result_slice]
    )


@pytest.mark.parametrize(
    "seg_list_slice,matcher_keywords,trim_noncode,result_slice",
    [
        # Matching the first element of the list
        (slice(None, None), ["bar"], False, slice(None, 1)),
        # Matching with a bit of whitespace before
        (slice(1, None), ["foo"], True, slice(1, 3)),
        # Matching with a bit of whitespace before (not trim_noncode)
        (slice(1, None), ["foo"], False, None),
        # Matching with whitespace after
        (slice(None, 2), ["bar"], True, slice(None, 2)),
    ],
)
def test__parser__grammar__base__longest_trimmed_match__basic(
    seg_list, seg_list_slice, matcher_keywords, trim_noncode, result_slice
):
    """Test the _longest_trimmed_match method of the BaseGrammar."""
    # Make the matcher keywords
    matchers = [StringParser(keyword, KeywordSegment) for keyword in matcher_keywords]

    with RootParseContext(dialect=None) as ctx:
        m, _ = BaseGrammar._longest_trimmed_match(
            seg_list[seg_list_slice], matchers, ctx, trim_noncode=trim_noncode
        )

    # Make the check tuple
    expected_result = make_result_tuple(
        result_slice=result_slice,
        matcher_keywords=matcher_keywords,
        seg_list=seg_list,
    )

    assert m.matched_segments == expected_result


def test__parser__grammar__base__longest_trimmed_match__adv(seg_list, caplog):
    """Test the _longest_trimmed_match method of the BaseGrammar."""
    bs = StringParser("bar", KeywordSegment)
    fs = StringParser("foo", KeywordSegment)
    matchers = [
        bs,
        fs,
        Sequence(bs, fs),  # This should be the winner.
        OneOf(bs, fs),
        Sequence(bs, fs),  # Another to check we return the first
    ]
    with RootParseContext(dialect=None) as ctx:
        # Matching the first element of the list
        with caplog.at_level(logging.DEBUG, logger="sqluff.parser"):
            match, matcher = BaseGrammar._longest_trimmed_match(seg_list, matchers, ctx)
    # Check we got a match
    assert match
    # Check we got the right one.
    assert matcher is matchers[2]
    # And it matched the first three segments
    assert len(match) == 3


@pytest.mark.parametrize(
    "seg_list_slice,matcher_keywords,result_slice,winning_matcher,pre_match_slice",
    [
        # Basic version, we should find bar first
        (slice(None, None), ["bar", "foo"], slice(None, 1), "bar", None),
        # Look ahead for foo
        (slice(None, None), ["foo"], slice(2, 3), "foo", slice(None, 2)),
    ],
)
def test__parser__grammar__base__look_ahead_match(
    seg_list_slice,
    matcher_keywords,
    result_slice,
    winning_matcher,
    pre_match_slice,
    seg_list,
):
    """Test the _look_ahead_match method of the BaseGrammar."""
    # Make the matcher keywords
    matchers = [StringParser(keyword, KeywordSegment) for keyword in matcher_keywords]
    # Fetch the matching keyword from above by index
    winning_matcher = matchers[matcher_keywords.index(winning_matcher)]

    with RootParseContext(dialect=None) as ctx:
        m = BaseGrammar._look_ahead_match(
            seg_list[seg_list_slice],
            matchers,
            ctx,
        )

    # Check structure of the response.
    assert isinstance(m, tuple)
    assert len(m) == 3
    # Unpack
    result_pre_match, result_match, result_matcher = m

    # Check the right matcher won
    assert result_matcher == winning_matcher

    # Make check tuple for the pre-match section
    if pre_match_slice:
        pre_match_slice = seg_list[pre_match_slice]
    else:
        pre_match_slice = ()
    assert result_pre_match == pre_match_slice

    # Make the check tuple
    expected_result = make_result_tuple(
        result_slice=result_slice,
        matcher_keywords=matcher_keywords,
        seg_list=seg_list,
    )
    assert result_match.matched_segments == expected_result


def test__parser__grammar__base__ephemeral_segment(seg_list):
    """Test the ephemeral features on BaseGrammar.

    Normally you can't call .match() on a BaseGrammar, but
    if things are set up right, then it should be possible
    in the case that the ephemeral_name is set.

    This indirectly tests the allow_ephemeral decorator.
    """
    g = BaseGrammar(ephemeral_name="TestGrammar")

    with RootParseContext(dialect=None) as ctx:
        m = g.match(seg_list, ctx)
        # Check we get an ephemeral segment
        assert isinstance(m.matched_segments[0], EphemeralSegment)
        assert len(m.matched_segments) == 1
        chkpoint = m.matched_segments[0]
        # Check it's got the same content.
        assert chkpoint.segments == seg_list


def test__parser__grammar__oneof__ephemeral_segment(seg_list):
    """A realistic full test of ephemeral segments."""

    class TestSegment(BaseSegment):
        match_grammar = OneOf(
            StringParser("bar", KeywordSegment), ephemeral_name="foofoo"
        )

    with RootParseContext(dialect=None) as ctx:
        m = TestSegment.match(seg_list[:1], ctx)
        # Make sure we've matched
        assert m
        seg = m.matched_segments[0]
        assert isinstance(seg, TestSegment)
        # Check the content is ephemeral
        assert isinstance(seg.segments[0], EphemeralSegment)
        assert seg.segments[0].ephemeral_name == "foofoo"
        # Expand the segment
        res = seg.parse(ctx)
        # Check we still have a test segment
        assert isinstance(res, TestSegment)
        # But that it contains a keyword segment now
        assert isinstance(res.segments[0], KeywordSegment)


def test__parser__grammar__base__bracket_sensitive_look_ahead_match(
    bracket_seg_list, fresh_ansi_dialect
):
    """Test the _bracket_sensitive_look_ahead_match method of the BaseGrammar."""
    bs = StringParser("bar", KeywordSegment)
    fs = StringParser("foo", KeywordSegment)
    # We need a dialect here to do bracket matching
    with RootParseContext(dialect=fresh_ansi_dialect) as ctx:
        # Basic version, we should find bar first
        pre_section, match, matcher = BaseGrammar._bracket_sensitive_look_ahead_match(
            bracket_seg_list, [fs, bs], ctx
        )
        assert pre_section == ()
        assert matcher == bs
        # NB the middle element is a match object
        assert match.matched_segments == (
            KeywordSegment("bar", bracket_seg_list[0].pos_marker),
        )

        # Look ahead for foo, we should find the one AFTER the brackets, not the
        # on IN the brackets.
        pre_section, match, matcher = BaseGrammar._bracket_sensitive_look_ahead_match(
            bracket_seg_list, [fs], ctx
        )
        # NB: The bracket segments will have been mutated, so we can't directly compare.
        # Make sure we've got a bracketed section in there.
        assert len(pre_section) == 5
        assert pre_section[2].is_type("bracketed")
        assert len(pre_section[2].segments) == 4
        assert matcher == fs
        # We shouldn't match the whitespace with the keyword
        assert match.matched_segments == (
            KeywordSegment("foo", bracket_seg_list[8].pos_marker),
        )
        # Check that the unmatched segments are nothing.
        assert not match.unmatched_segments


def test__parser__grammar__base__bracket_fail_with_open_paren_close_square_mismatch(
    generate_test_segments, fresh_ansi_dialect
):
    """Test _bracket_sensitive_look_ahead_match failure case.

    Should fail when the type of a close bracket doesn't match the type of the
    corresponding open bracket, but both are "definite" brackets.
    """
    fs = StringParser("foo", KeywordSegment)
    # We need a dialect here to do bracket matching
    with RootParseContext(dialect=fresh_ansi_dialect) as ctx:
        # Basic version, we should find bar first
        with pytest.raises(SQLParseError) as sql_parse_error:
            BaseGrammar._bracket_sensitive_look_ahead_match(
                generate_test_segments(
                    [
                        "select",
                        " ",
                        "*",
                        " ",
                        "from",
                        "(",
                        "foo",
                        "]",  # Bracket types don't match (parens vs square)
                    ]
                ),
                [fs],
                ctx,
            )
        assert sql_parse_error.match("Found unexpected end bracket")


def test__parser__grammar__base__bracket_fail_with_unexpected_end_bracket(
    generate_test_segments, fresh_ansi_dialect
):
    """Test _bracket_sensitive_look_ahead_match edge case.

    Should fail gracefully and stop matching if we find a trailing unmatched.
    """
    fs = StringParser("foo", KeywordSegment)
    # We need a dialect here to do bracket matching
    with RootParseContext(dialect=fresh_ansi_dialect) as ctx:
        _, match, _ = BaseGrammar._bracket_sensitive_look_ahead_match(
            generate_test_segments(
                [
                    "bar",
                    "(",  # This bracket pair should be mutated
                    ")",
                    " ",
                    ")",  # This is the unmatched bracket
                    " ",
                    "foo",
                ]
            ),
            [fs],
            ctx,
        )
        # Check we don't match (even though there's a foo at the end)
        assert not match
        # Check the first bracket pair have been mutated.
        segs = match.unmatched_segments
        assert segs[1].is_type("bracketed")
        assert segs[1].raw == "()"
        assert len(segs[1].segments) == 2
        # Check the trailing foo hasn't been mutated
        assert segs[5].raw == "foo"
        assert not isinstance(segs[5], KeywordSegment)


def test__parser__grammar__ref_eq():
    """Test equality of Ref Grammars."""
    r1 = Ref("foo")
    r2 = Ref("foo")
    assert r1 is not r2
    assert r1 == r2
    check_list = [1, 2, r2, 3]
    # Check we can find it in lists
    assert r1 in check_list
    # Check we can get it's position
    assert check_list.index(r1) == 2
    # Check we can remove it from a list
    check_list.remove(r1)
    assert r1 not in check_list


def test__parser__grammar_ref_exclude(generate_test_segments, fresh_ansi_dialect):
    """Test the Ref grammar exclude option."""
    ni = Ref("NakedIdentifierSegment", exclude=Ref.keyword("ABS"))
    ts = generate_test_segments(["ABS", "ABSOLUTE"])
    with RootParseContext(dialect=fresh_ansi_dialect) as ctx:
        # Asset ABS does not match, due to the exclude
        assert not ni.match([ts[0]], parse_context=ctx)
        # Asset ABSOLUTE does match
        assert ni.match([ts[1]], parse_context=ctx)


def test__parser__grammar__oneof__copy():
    """Test grammar copying."""
    bs = StringParser("bar", KeywordSegment)
    fs = StringParser("foo", KeywordSegment)
    g1 = OneOf(fs, bs)
    # Check copy
    g2 = g1.copy()
    assert g1 == g2
    assert g1 is not g2
    # Check copy insert (start)
    g3 = g1.copy(insert=[bs], at=0)
    assert g3 == OneOf(bs, fs, bs)
    # Check copy insert (mid)
    g4 = g1.copy(insert=[bs], at=1)
    assert g4 == OneOf(fs, bs, bs)
    # Check copy insert (end)
    g5 = g1.copy(insert=[bs], at=-1)
    assert g5 == OneOf(fs, bs, bs)


@pytest.mark.parametrize("allow_gaps", [True, False])
def test__parser__grammar_oneof(seg_list, allow_gaps):
    """Test the OneOf grammar.

    NB: Should behave the same regardless of code_only.

    """
    bs = StringParser("bar", KeywordSegment)
    fs = StringParser("foo", KeywordSegment)
    g = OneOf(fs, bs, allow_gaps=allow_gaps)
    with RootParseContext(dialect=None) as ctx:
        # Check directly
        assert g.match(seg_list, parse_context=ctx).matched_segments == (
            KeywordSegment("bar", seg_list[0].pos_marker),
        )
        # Check with a bit of whitespace
        assert not g.match(seg_list[1:], parse_context=ctx)


def test__parser__grammar_oneof_templated(seg_list):
    """Test the OneOf grammar.

    NB: Should behave the same regardless of code_only.

    """
    bs = StringParser("bar", KeywordSegment)
    fs = StringParser("foo", KeywordSegment)
    g = OneOf(fs, bs)
    with RootParseContext(dialect=None) as ctx:
        # This shouldn't match, but it *ALSO* shouldn't raise an exception.
        # https://github.com/sqlfluff/sqlfluff/issues/780
        assert not g.match(seg_list[5:], parse_context=ctx)


def test__parser__grammar_oneof_exclude(seg_list):
    """Test the OneOf grammar exclude option."""
    bs = StringParser("bar", KeywordSegment)
    fs = StringParser("foo", KeywordSegment)
    g = OneOf(bs, exclude=Sequence(bs, fs))
    with RootParseContext(dialect=None) as ctx:
        # Just against the first alone
        assert g.match(seg_list[:1], parse_context=ctx)
        # Now with the bit to exclude included
        assert not g.match(seg_list, parse_context=ctx)


def test__parser__grammar_oneof_take_longest_match(seg_list):
    """Test that the OneOf grammar takes the longest match."""
    fooRegex = RegexParser(r"fo{2}", KeywordSegment)
    baar = StringParser("baar", KeywordSegment)
    foo = StringParser("foo", KeywordSegment)
    fooBaar = Sequence(
        foo,
        baar,
    )

    # Even if fooRegex comes first, fooBaar
    # is a longer match and should be taken
    g = OneOf(fooRegex, fooBaar)
    with RootParseContext(dialect=None) as ctx:
        assert fooRegex.match(seg_list[2:], parse_context=ctx).matched_segments == (
            KeywordSegment("foo", seg_list[2].pos_marker),
        )
        assert g.match(seg_list[2:], parse_context=ctx).matched_segments == (
            KeywordSegment("foo", seg_list[2].pos_marker),
            KeywordSegment("baar", seg_list[3].pos_marker),
        )


def test__parser__grammar_oneof_take_first(seg_list):
    """Test that the OneOf grammar takes first match in case they are of same length."""
    fooRegex = RegexParser(r"fo{2}", KeywordSegment)
    foo = StringParser("foo", KeywordSegment)

    # Both segments would match "foo"
    # so we test that order matters
    g1 = OneOf(fooRegex, foo)
    g2 = OneOf(foo, fooRegex)
    with RootParseContext(dialect=None) as ctx:
        assert g1.match(seg_list[2:], parse_context=ctx).matched_segments == (
            KeywordSegment("foo", seg_list[2].pos_marker),
        )
        assert g2.match(seg_list[2:], parse_context=ctx).matched_segments == (
            KeywordSegment("foo", seg_list[2].pos_marker),
        )


def test__parser__grammar_startswith_a():
    """Test the StartsWith grammar fails when no terminator supplied."""
    Keyword = StringParser("foo", KeywordSegment)
    with pytest.raises(AssertionError):
        StartsWith(Keyword)


@pytest.mark.parametrize(
    "include_terminator,match_length",
    [
        (False, 3),
        # NB: In this case we still shouldn't match the trailing whitespace.
        (True, 4),
    ],
)
def test__parser__grammar_startswith_b(
    include_terminator, match_length, seg_list, fresh_ansi_dialect, caplog
):
    """Test the StartsWith grammar with a terminator (included & excluded)."""
    baar = StringParser("baar", KeywordSegment)
    bar = StringParser("bar", KeywordSegment)
    grammar = StartsWith(bar, terminator=baar, include_terminator=include_terminator)
    with RootParseContext(dialect=fresh_ansi_dialect) as ctx:
        with caplog.at_level(logging.DEBUG, logger="sqlfluff.parser"):
            m = grammar.match(seg_list, parse_context=ctx)
            assert len(m) == match_length


def test__parser__grammar_sequence(seg_list, caplog):
    """Test the Sequence grammar."""
    bs = StringParser("bar", KeywordSegment)
    fs = StringParser("foo", KeywordSegment)
    g = Sequence(bs, fs)
    # If running in the test environment, assert that Sequence recognises this
    if getenv("SQLFLUFF_TESTENV", ""):
        assert g.test_env
    gc = Sequence(bs, fs, allow_gaps=False)
    with RootParseContext(dialect=None) as ctx:
        with caplog.at_level(logging.DEBUG, logger="sqlfluff.parser"):
            # Should be able to match the list using the normal matcher
            logging.info("#### TEST 1")
            m = g.match(seg_list, parse_context=ctx)
            assert m
            assert len(m) == 3
            assert m.matched_segments == (
                KeywordSegment("bar", seg_list[0].pos_marker),
                seg_list[1],  # This will be the whitespace segment
                KeywordSegment("foo", seg_list[2].pos_marker),
            )
            # Shouldn't with the allow_gaps matcher
            logging.info("#### TEST 2")
            assert not gc.match(seg_list, parse_context=ctx)
            # Shouldn't match even on the normal one if we don't start at the beginning
            logging.info("#### TEST 2")
            assert not g.match(seg_list[1:], parse_context=ctx)


def test__parser__grammar_sequence_nested(seg_list, caplog):
    """Test the Sequence grammar when nested."""
    bs = StringParser("bar", KeywordSegment)
    fs = StringParser("foo", KeywordSegment)
    bas = StringParser("baar", KeywordSegment)
    g = Sequence(Sequence(bs, fs), bas)
    with RootParseContext(dialect=None) as ctx:
        with caplog.at_level(logging.DEBUG, logger="sqlfluff.parser"):
            # Matching the start of the list shouldn't work
            logging.info("#### TEST 1")
            assert not g.match(seg_list[:2], parse_context=ctx)
            # Matching the whole list should, and the result should be flat
            logging.info("#### TEST 2")
            assert g.match(seg_list, parse_context=ctx).matched_segments == (
                KeywordSegment("bar", seg_list[0].pos_marker),
                seg_list[1],  # This will be the whitespace segment
                KeywordSegment("foo", seg_list[2].pos_marker),
                KeywordSegment("baar", seg_list[3].pos_marker)
                # NB: No whitespace at the end, this shouldn't be consumed.
            )


def test__parser__grammar_sequence_indent(seg_list, caplog):
    """Test the Sequence grammar with indents."""
    bs = StringParser("bar", KeywordSegment)
    fs = StringParser("foo", KeywordSegment)
    g = Sequence(Indent, bs, fs)
    with RootParseContext(dialect=None) as ctx:
        with caplog.at_level(logging.DEBUG, logger="sqlfluff.parser"):
            m = g.match(seg_list, parse_context=ctx)
            assert m
            # check we get an indent.
            assert isinstance(m.matched_segments[0], Indent)
            assert isinstance(m.matched_segments[1], KeywordSegment)


def test__parser__grammar_sequence_indent_conditional(seg_list, caplog):
    """Test the Sequence grammar with indents."""
    bs = StringParser("bar", KeywordSegment)
    fs = StringParser("foo", KeywordSegment)
    # We will assume the default config has indented_joins = False.
    # We're testing without explicitly setting the `config_type` because
    # that's the assumed way of using the grammar in practice.
    g = Sequence(
        Conditional(Indent, indented_joins=False),
        bs,
        Conditional(Indent, indented_joins=True),
        fs,
    )
    with RootParseContext(dialect=None) as ctx:
        with caplog.at_level(logging.DEBUG, logger="sqlfluff.parser"):
            m = g.match(seg_list, parse_context=ctx)
            assert m
            # Check we get an Indent.
            assert isinstance(m.matched_segments[0], Indent)
            assert isinstance(m.matched_segments[1], KeywordSegment)
            # check the whitespace is still there
            assert isinstance(m.matched_segments[2], WhitespaceSegment)
            # Check the second Indent does not appear
            assert not isinstance(m.matched_segments[3], Indent)
            assert isinstance(m.matched_segments[3], KeywordSegment)


@pytest.mark.parametrize(
    "token_list,min_delimiters,allow_gaps,allow_trailing,match_len",
    [
        # Basic testing
        (["bar", " \t ", ".", "    ", "bar"], None, True, False, 5),
        (["bar", " \t ", ".", "    ", "bar", "    "], None, True, False, 6),
        # Testing allow_trailing
        (["bar", " \t ", ".", "   "], None, True, False, 0),
        (["bar", " \t ", ".", "   "], None, True, True, 4),
        # Testing the implications of allow_gaps
        (["bar", " \t ", ".", "    ", "bar"], 0, True, False, 5),
        (["bar", " \t ", ".", "    ", "bar"], 0, False, False, 1),
        (["bar", " \t ", ".", "    ", "bar"], 1, True, False, 5),
        (["bar", " \t ", ".", "    ", "bar"], 1, False, False, 0),
        (["bar", ".", "bar"], None, True, False, 3),
        (["bar", ".", "bar"], None, False, False, 3),
        (["bar", ".", "bar"], 1, True, False, 3),
        (["bar", ".", "bar"], 1, False, False, 3),
        # Check we still succeed with something trailing right on the end.
        (["bar", ".", "bar", "foo"], 1, False, False, 3),
        # Check min_delimiters. There's a delimiter here, but not enough to match.
        (["bar", ".", "bar", "foo"], 2, True, False, 0),
    ],
)
def test__parser__grammar_delimited(
    min_delimiters,
    allow_gaps,
    allow_trailing,
    token_list,
    match_len,
    caplog,
    generate_test_segments,
    fresh_ansi_dialect,
):
    """Test the Delimited grammar when not code_only."""
    seg_list = generate_test_segments(token_list)
    g = Delimited(
        StringParser("bar", KeywordSegment),
        delimiter=StringParser(".", SymbolSegment),
        allow_gaps=allow_gaps,
        allow_trailing=allow_trailing,
        min_delimiters=min_delimiters,
    )
    with RootParseContext(dialect=fresh_ansi_dialect) as ctx:
        with caplog.at_level(logging.DEBUG, logger="sqlfluff.parser"):
            # Matching with whitespace shouldn't match if we need at least one delimiter
            m = g.match(seg_list, parse_context=ctx)
            assert len(m) == match_len


@pytest.mark.parametrize(
    "keyword,enforce_ws,slice_len",
    [
        # Basic testing
        ("foo", False, 1),
        # Greedy matching until the first item should return none
        ("bar", False, 0),
        # Greedy matching up to baar should return bar, foo...
        ("baar", False, 3),
        # ... except if whitespace is required to precede it
        ("baar", True, 6),
    ],
)
def test__parser__grammar_greedyuntil(
    keyword, seg_list, enforce_ws, slice_len, fresh_ansi_dialect
):
    """Test the GreedyUntil grammar."""
    grammar = GreedyUntil(
        StringParser(keyword, KeywordSegment),
        enforce_whitespace_preceding_terminator=enforce_ws,
    )
    with RootParseContext(dialect=fresh_ansi_dialect) as ctx:
        assert (
            grammar.match(seg_list, parse_context=ctx).matched_segments
            == seg_list[:slice_len]
        )


def test__parser__grammar_greedyuntil_bracketed(bracket_seg_list, fresh_ansi_dialect):
    """Test the GreedyUntil grammar with brackets."""
    fs = StringParser("foo", KeywordSegment)
    g = GreedyUntil(fs)
    with RootParseContext(dialect=fresh_ansi_dialect) as ctx:
        # Check that we can make it past the brackets
        match = g.match(bracket_seg_list, parse_context=ctx)
        assert len(match) == 4
        # Check we successfully constructed a bracketed segment
        assert match.matched_segments[2].is_type("bracketed")
        assert match.matched_segments[2].raw == "(foo    )"
        # Check that the unmatched segments is foo AND the whitespace
        assert len(match.unmatched_segments) == 2


def test__parser__grammar_anything(seg_list, fresh_ansi_dialect):
    """Test the Anything grammar."""
    with RootParseContext(dialect=fresh_ansi_dialect) as ctx:
        assert Anything().match(seg_list, parse_context=ctx)


def test__parser__grammar_nothing(seg_list, fresh_ansi_dialect):
    """Test the Nothing grammar."""
    with RootParseContext(dialect=fresh_ansi_dialect) as ctx:
        assert not Nothing().match(seg_list, parse_context=ctx)


def test__parser__grammar_noncode(seg_list, fresh_ansi_dialect):
    """Test the NonCodeMatcher."""
    with RootParseContext(dialect=fresh_ansi_dialect) as ctx:
        m = NonCodeMatcher().match(seg_list[1:], parse_context=ctx)
        # NonCode Matcher doesn't work with simple
        assert NonCodeMatcher().simple(ctx) is None
    # We should match one and only one segment
    assert len(m) == 1


def test__parser__grammar_anysetof(generate_test_segments):
    """Test the AnySetOf grammar."""
    token_list = ["bar", "  \t ", "foo", "  \t ", "bar"]
    seg_list = generate_test_segments(token_list)

    bs = StringParser("bar", KeywordSegment)
    fs = StringParser("foo", KeywordSegment)
    g = AnySetOf(fs, bs)
    with RootParseContext(dialect=None) as ctx:
        # Check directly
        assert g.match(seg_list, parse_context=ctx).matched_segments == (
            KeywordSegment("bar", seg_list[0].pos_marker),
            WhitespaceSegment("  \t ", seg_list[1].pos_marker),
            KeywordSegment("foo", seg_list[2].pos_marker),
        )
        # Check with a bit of whitespace
        assert not g.match(seg_list[1:], parse_context=ctx)



"""Implementation of Rule L054."""
from typing import Optional, List

from sqlfluff.core.rules import BaseRule, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.core.rules.doc_decorators import document_configuration, document_groups
from sqlfluff.utils.functional import sp, FunctionalContext


@document_groups
@document_configuration
class Rule_L054(BaseRule):
    """Inconsistent column references in ``GROUP BY/ORDER BY`` clauses.

    .. note::
       ``ORDER BY`` clauses from ``WINDOW`` clauses are ignored by this rule.

    **Anti-pattern**

    A mix of implicit and explicit column references are used in a ``GROUP BY``
    clause.

    .. code-block:: sql
       :force:

        SELECT
            foo,
            bar,
            sum(baz) AS sum_value
        FROM fake_table
        GROUP BY
            foo, 2;

        -- The same also applies to column
        -- references in ORDER BY clauses.

        SELECT
            foo,
            bar
        FROM fake_table
        ORDER BY
            1, bar;

    **Best practice**

    Reference all ``GROUP BY``/``ORDER BY`` columns either by name or by position.

    .. code-block:: sql
       :force:

        -- GROUP BY: Explicit
        SELECT
            foo,
            bar,
            sum(baz) AS sum_value
        FROM fake_table
        GROUP BY
            foo, bar;

        -- ORDER BY: Explicit
        SELECT
            foo,
            bar
        FROM fake_table
        ORDER BY
            foo, bar;

        -- GROUP BY: Implicit
        SELECT
            foo,
            bar,
            sum(baz) AS sum_value
        FROM fake_table
        GROUP BY
            1, 2;

        -- ORDER BY: Implicit
        SELECT
            foo,
            bar
        FROM fake_table
        ORDER BY
            1, 2;
    """

    groups = ("all", "core")
    config_keywords = ["group_by_and_order_by_style"]
    crawl_behaviour = SegmentSeekerCrawler({"groupby_clause", "orderby_clause"})
    _ignore_types: List[str] = ["withingroup_clause", "window_specification"]

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Inconsistent column references in GROUP BY/ORDER BY clauses."""
        # Config type hints
        self.group_by_and_order_by_style: str

        # We only care about GROUP BY/ORDER BY clauses.
        assert context.segment.is_type("groupby_clause", "orderby_clause")

        # Ignore Windowing clauses
        if FunctionalContext(context).parent_stack.any(sp.is_type(*self._ignore_types)):
            return LintResult(memory=context.memory)

        # Look at child segments and map column references to either the implicit or
        # explicit category.
        # N.B. segment names are used as the numeric literal type is 'raw', so best to
        # be specific with the name.
        column_reference_category_map = {
            "column_reference": "explicit",
            "expression": "explicit",
            "numeric_literal": "implicit",
        }
        column_reference_category_set = {
            column_reference_category_map[segment.get_type()]
            for segment in context.segment.segments
            if segment.is_type(*column_reference_category_map.keys())
        }

        # If there are no column references then just return
        if not column_reference_category_set:  # pragma: no cover
            return LintResult(memory=context.memory)

        if self.group_by_and_order_by_style == "consistent":
            # If consistent naming then raise lint error if either:

            if len(column_reference_category_set) > 1:
                # 1. Both implicit and explicit column references are found in the same
                # clause.
                return LintResult(
                    anchor=context.segment,
                    memory=context.memory,
                )
            else:
                # 2. A clause is found to contain column name references that
                #    contradict the precedent set in earlier clauses.
                current_group_by_order_by_convention = (
                    column_reference_category_set.pop()
                )
                prior_group_by_order_by_convention = context.memory.get(
                    "prior_group_by_order_by_convention"
                )

                if prior_group_by_order_by_convention and (
                    prior_group_by_order_by_convention
                    != current_group_by_order_by_convention
                ):
                    return LintResult(
                        anchor=context.segment,
                        memory=context.memory,
                    )

                context.memory[
                    "prior_group_by_order_by_convention"
                ] = current_group_by_order_by_convention
        else:
            # If explicit or implicit naming then raise lint error
            # if the opposite reference type is detected.
            if any(
                category != self.group_by_and_order_by_style
                for category in column_reference_category_set
            ):
                return LintResult(
                    anchor=context.segment,
                    memory=context.memory,
                )

        # Return memory for later clauses.
        return LintResult(memory=context.memory)



"""Implementation of Rule L055."""
from typing import Optional

from sqlfluff.core.rules import BaseRule, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.core.rules.doc_decorators import document_groups


@document_groups
class Rule_L055(BaseRule):
    """Use ``LEFT JOIN`` instead of ``RIGHT JOIN``.

    **Anti-pattern**

    ``RIGHT JOIN`` is used.

    .. code-block:: sql
       :force:

        SELECT
            foo.col1,
            bar.col2
        FROM foo
        RIGHT JOIN bar
            ON foo.bar_id = bar.id;

    **Best practice**

    Refactor and use ``LEFT JOIN`` instead.

    .. code-block:: sql
       :force:

        SELECT
            foo.col1,
            bar.col2
        FROM bar
        LEFT JOIN foo
            ON foo.bar_id = bar.id;
    """

    groups = ("all",)
    crawl_behaviour = SegmentSeekerCrawler({"join_clause"})

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Use LEFT JOIN instead of RIGHT JOIN."""
        # We are only interested in JOIN clauses.
        assert context.segment.is_type("join_clause")

        # Identify if RIGHT JOIN is present.
        if {"RIGHT", "JOIN"}.issubset(
            {segment.raw_upper for segment in context.segment.segments}
        ):
            return LintResult(context.segment.segments[0])

        return None



"""Tests specific to the exasol dialect."""
import pytest

TEST_DIALECT = "exasol"


# Develop test to check specific elements against specific grammars.
@pytest.mark.parametrize(
    "segmentref,raw",
    [
        ("RangeOperator", ".."),
        ("WalrusOperatorSegment", ":="),
        ("VariableNameSegment", "var1"),
    ],
)
def test_dialect_exasol_specific_segment_parses(
    segmentref, raw, caplog, dialect_specific_segment_parses
):
    """Test exasol specific segments."""
    dialect_specific_segment_parses(TEST_DIALECT, segmentref, raw, caplog)



"""Errors - these are closely linked to what used to be called violations."""
from typing import Optional, Tuple, Any, List

CheckTuple = Tuple[str, int, int]


class SQLBaseError(ValueError):
    """Base Error Class for all violations."""

    _code: Optional[str] = None
    _identifier = "base"

    def __init__(
        self,
        *args,
        pos=None,
        line_no=0,
        line_pos=0,
        ignore=False,
        fatal=False,
        warning=False,
        **kwargs
    ):
        self.fatal = fatal
        self.ignore = ignore
        self.warning = warning
        if pos:
            self.line_no, self.line_pos = pos.source_position()
        else:
            self.line_no = line_no
            self.line_pos = line_pos
        super().__init__(*args, **kwargs)

    @property
    def fixable(self):
        """Should this error be considered fixable?"""
        return False

    def rule_code(self) -> str:
        """Fetch the code of the rule which cause this error.

        NB: This only returns a real code for some subclasses of
        error, (the ones with a `rule` attribute), but otherwise
        returns a placeholder value which can be used instead.
        """
        if hasattr(self, "rule"):
            return getattr(self, "rule").code

        return self._code or "????"

    def desc(self) -> str:
        """Fetch a description of this violation.

        NB: For violations which don't directly implement a rule
        this attempts to return the error message linked to whatever
        caused the violation. Optionally some errors may have their
        description set directly.
        """
        if hasattr(self, "description") and getattr(self, "description", None):
            # This can only override if it's present AND
            # if it's non-null.
            return getattr(self, "description")

        if hasattr(self, "rule"):
            return getattr(self, "rule").description

        # Return the first element - probably a string message
        if len(self.args) > 1 and isinstance(self.args, str):  # pragma: no cover TODO?
            return self.args

        if len(self.args) == 1:
            return self.args[0]

        return self.__class__.__name__  # pragma: no cover

    def get_info_dict(self):
        """Return a dict of properties.

        This is useful in the API for outputting violations.
        """
        return {
            "line_no": self.line_no,
            "line_pos": self.line_pos,
            "code": self.rule_code(),
            "description": self.desc(),
        }

    def check_tuple(self) -> CheckTuple:
        """Get a tuple representing this error. Mostly for testing."""
        return (
            self.rule_code(),
            self.line_no,
            self.line_pos,
        )

    def source_signature(self) -> Tuple[Any, ...]:
        """Return hashable source signature for deduplication."""
        return (self.check_tuple(), self.desc())

    def ignore_if_in(self, ignore_iterable: List[str]):
        """Ignore this violation if it matches the iterable."""
        if self._identifier in ignore_iterable:
            self.ignore = True

    def warning_if_in(self, warning_iterable: List[str]):
        """Warning only for this violation if it matches the iterable.

        Designed for rule codes so works with L001, L00X but also TMP or PRS
        for templating and parsing errors.
        """
        if self.rule_code() in warning_iterable:
            self.warning = True


class SQLTemplaterError(SQLBaseError):
    """An error which occurred during templating.

    Args:
        pos (:obj:`PosMarker`, optional): The position which the error
            occurred at.

    """

    _code = "TMP"
    _identifier = "templating"


class SQLFluffSkipFile(RuntimeError):
    """An error returned from a templater to skip a file."""

    pass


class SQLLexError(SQLBaseError):
    """An error which occurred during lexing.

    Args:
        pos (:obj:`PosMarker`, optional): The position which the error
            occurred at.

    """

    _code = "LXR"
    _identifier = "lexing"


class SQLParseError(SQLBaseError):
    """An error which occurred during parsing.

    Args:
        segment (:obj:`BaseSegment`, optional): The segment which is relevant
            for the failure in parsing. This is likely to be a subclass of
            `BaseSegment` rather than the parent class itself. This is mostly
            used for logging and for referencing position.

    """

    _code = "PRS"
    _identifier = "parsing"

    def __init__(self, *args, segment=None, **kwargs):
        # Store the segment on creation - we might need it later
        self.segment = segment
        if self.segment:
            kwargs["pos"] = self.segment.pos_marker
        super().__init__(*args, **kwargs)


class SQLLintError(SQLBaseError):
    """An error which occurred during linting.

    In particular we reference the rule here to do extended logging based on
    the rule in question which caused the fail.

    Args:
        segment (:obj:`BaseSegment`, optional): The segment which is relevant
            for the failure in parsing. This is likely to be a subclass of
            `BaseSegment` rather than the parent class itself. This is mostly
            used for logging and for referencing position.

    """

    _identifier = "linting"

    def __init__(
        self, *args, segment=None, rule=None, fixes=None, description=None, **kwargs
    ):
        # Something about position, message and fix?
        self.segment = segment
        if self.segment:
            kwargs["pos"] = self.segment.pos_marker
        self.rule = rule
        self.fixes = fixes or []
        self.description = description
        super().__init__(*args, **kwargs)

    @property
    def fixable(self):
        """Should this error be considered fixable?"""
        if self.fixes:
            return True
        return False

    def source_signature(self) -> Tuple[Any, ...]:
        """Return hashable source signature for deduplication.

        For linting errors we need to dedupe on more than just location and
        description, we also need to check the edits potentially made, both
        in the templated file but also in the source.
        """
        fix_raws = tuple(
            tuple(e.raw for e in f.edit) if f.edit else None for f in self.fixes
        )
        source_fixes = tuple(
            tuple(tuple(e.source_fixes) for e in f.edit) if f.edit else None
            for f in self.fixes
        )
        return (self.check_tuple(), self.description, fix_raws, source_fixes)

    def __repr__(self):
        return "<SQLLintError: rule {} pos:{!r}, #fixes: {}, description: {}>".format(
            self.rule_code(),
            (self.line_no, self.line_pos),
            len(self.fixes),
            self.description,
        )


class SQLFluffUserError(ValueError):
    """An error which should be fed back to the user."""



"""Implementation of Rule L033."""
from sqlfluff.core.parser import (
    WhitespaceSegment,
    KeywordSegment,
)

from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.core.rules.doc_decorators import document_groups


@document_groups
class Rule_L033(BaseRule):
    """``UNION [DISTINCT|ALL]`` is preferred over just ``UNION``.

    .. note::
       This rule is only enabled for dialects that support ``UNION`` and
       ``UNION DISTINCT`` (``ansi``, ``hive``, ``mysql``, and ``redshift``).

    **Anti-pattern**

    In this example, ``UNION DISTINCT`` should be preferred over ``UNION``, because
    explicit is better than implicit.

    .. code-block:: sql

        SELECT a, b FROM table_1
        UNION
        SELECT a, b FROM table_2

    **Best practice**

    Specify ``DISTINCT`` or ``ALL`` after ``UNION`` (note that ``DISTINCT`` is the
    default behavior).

    .. code-block:: sql

        SELECT a, b FROM table_1
        UNION DISTINCT
        SELECT a, b FROM table_2

    """

    groups = ("all", "core")
    crawl_behaviour = SegmentSeekerCrawler({"set_operator"})

    def _eval(self, context: RuleContext) -> LintResult:
        """Look for UNION keyword not immediately followed by DISTINCT or ALL.

        Note that UNION DISTINCT is valid, rule only applies to bare UNION.
        The function does this by looking for a segment of type set_operator
        which has a UNION but no DISTINCT or ALL.

        Note only some dialects have concept of UNION DISTINCT, so rule is only
        applied to dialects that are known to support this syntax.
        """
        if context.dialect.name not in [
            "ansi",
            "hive",
            "mysql",
            "redshift",
        ]:
            return LintResult()

        assert context.segment.is_type("set_operator")
        if "union" in context.segment.raw and not (
            "ALL" in context.segment.raw.upper()
            or "DISTINCT" in context.segment.raw.upper()
        ):
            return LintResult(
                anchor=context.segment,
                fixes=[
                    LintFix.replace(
                        context.segment.segments[0],
                        [
                            KeywordSegment("union"),
                            WhitespaceSegment(),
                            KeywordSegment("distinct"),
                        ],
                    )
                ],
            )
        elif "UNION" in context.segment.raw.upper() and not (
            "ALL" in context.segment.raw.upper()
            or "DISTINCT" in context.segment.raw.upper()
        ):
            return LintResult(
                anchor=context.segment,
                fixes=[
                    LintFix.replace(
                        context.segment.segments[0],
                        [
                            KeywordSegment("UNION"),
                            WhitespaceSegment(),
                            KeywordSegment("DISTINCT"),
                        ],
                    )
                ],
            )
        return LintResult()



"""A list of all Athena keywords.

Presto List: https://prestodb.io/docs/0.217/language/reserved.html
Hive List: https://cwiki.apache.org/confluence/display/Hive/LanguageManual+DDL
"""

athena_reserved_keywords = [
    "ALL",
    "ALTER",
    "AND",
    "ARRAY",
    "AS",
    "AUTHORIZATION",
    "BETWEEN",
    "BIGINT",
    "BINARY",
    "BOOLEAN",
    "BOTH",
    "BY",
    "CACHE",
    "CASE",
    "CAST",
    "CHAR",
    "COLUMN",
    "COMMIT",
    "CONF",
    "CONSTRAINT",
    "CREATE",
    "CROSS",
    "CUBE",
    "CURRENT_DATE",
    "CURRENT_TIMESTAMP",
    "CURRENT",
    "CURSOR",
    "DATABASE",
    "DATE",
    "DAYOFWEEK",
    "DECIMAL",
    "DELETE",
    "DESCRIBE",
    "DISTINCT",
    "DOUBLE",
    "DROP",
    "ELSE",
    "END",
    "EXCHANGE",
    "EXISTS",
    "EXTENDED",
    "EXTERNAL",
    "EXTRACT",
    "FALSE",
    "FETCH",
    "FLOAT",
    "FLOOR",
    "FOLLOWING",
    "FOR",
    "FOREIGN",
    "FROM",
    "FULL",
    "FUNCTION",
    "GRANT",
    "GROUP",
    "GROUPING",
    "HAVING",
    "IF",
    "IMPORT",
    "IN",
    "INNER",
    "INSERT",
    "INT",
    "INTEGER",
    "INTERSECT",
    "INTERVAL",
    "INTO",
    "IS",
    "JOIN",
    "LATERAL",
    "LEFT",
    "LESS",
    "LIKE",
    "LOCAL",
    "MACRO",
    "MAP",
    "MORE",
    "NONE",
    "NOT",
    "NULL",
    "NUMERIC",
    "OF",
    "ON",
    "ONLY",
    "OR",
    "ORDER",
    "OUT",
    "OUTER",
    "OVER",
    "PARTIALSCAN",
    "PARTITION",
    "PERCENT",
    "PRECEDING",
    "PRECISION",
    "PRESERVE",
    "PRIMARY",
    "PROCEDURE",
    "RANGE",
    "READS",
    "REDUCE",
    "REFERENCES",
    "REGEXP",
    "REVOKE",
    "RIGHT",
    "RLIKE",
    "ROLLBACK",
    "ROLLUP",
    "ROW",
    "ROWS",
    "SELECT",
    "SET",
    "SMALLINT",
    "START",
    "SYNC",
    "TABLE",
    "TABLESAMPLE",
    "THEN",
    "TIME",
    "TIMESTAMP",
    "TO",
    "TRANSFORM",
    "TRIGGER",
    "TRUE",
    "TRUNCATE",
    "UNBOUNDED",
    "UNION",
    "UNIQUEJOIN",
    "UPDATE",
    "USER",
    "USING",
    "UTC_TMESTAMP",
    "VALUES",
    "VARCHAR",
    "VIEWS",
    "WHEN",
    "WHERE",
    "WINDOW",
    "WITH",
]

athena_unreserved_keywords = [
    "ABORT",
    "ADD",
    "ADMIN",
    "AFTER",
    "ANALYZE",
    "ARCHIVE",
    "ASC",
    "AUTOCOMMIT",
    "BEFORE",
    "BUCKET_COUNT",
    "BUCKET",
    "BUCKETED_BY",
    "BUCKETS",
    "CASCADE",
    "CHANGE",
    "CLUSTER",
    "CLUSTERED",
    "CLUSTERSTATUS",
    "COLLECTION",
    "COLUMNS",
    "COMMENT",
    "COMPACT",
    "COMPACTIONS",
    "COMPRESSION",
    "COMPUTE",
    "CONCATENATE",
    "CONTINUE",
    "DATA",
    "DATABASES",
    "DATETIME",
    "DAY",
    "DBPROPERTIES",
    "DEFERRED",
    "DEFINED",
    "DELIMITED",
    "DEPENDENCY",
    "DESC",
    "DIRECTORIES",
    "DIRECTORY",
    "DISABLE",
    "DISTRIBUTE",
    "ELEM_TYPE",
    "ENABLE",
    "ESCAPED",
    "EXCLUSIVE",
    "EXPLAIN",
    "EXPORT",
    "EXTERNAL_LOCATION",
    "FIELD_DELIMITER",
    "FIELDS",
    "FILE",
    "FILEFORMAT",
    "FIRST",
    "FORMAT",
    "FORMAT",
    "FORMATTED",
    "FUNCTIONS",
    "HOLD_DDLTIME",
    "HOUR",
    "HYPERLOGLOG",
    "IDXPROPERTIES",
    "IGNORE",
    "INDEX",
    "INDEXES",
    "INPATH",
    "INPUTDRIVER",
    "INPUTFORMAT",
    "IPADDRESS",
    "ISOLATION",
    "ITEMS",
    "JAR",
    "KEY_TYPE",
    "KEY",
    "KEYS",
    "LAST",
    "LEVEL",
    "LIMIT",
    "LINES",
    "LOAD",
    "LOCATION",
    "LOCK",
    "LOCKS",
    "LOGICAL",
    "LONG",
    "MAPJOIN",
    "MATERIALIZED",
    "METADATA",
    "MINUS",
    "MINUTE",
    "MONTH",
    "MSCK",
    "NO_DROP",
    "NORELY",
    "NOSCAN",
    "NOVALIDATE",
    "NULLS",
    "OFFLINE",
    "OFFSET",
    "OPTION",
    "ORC_COMPRESSION",
    "OUTPUTDRIVER",
    "OUTPUTFORMAT",
    "OVERWRITE",
    "OWNER",
    "P4HYPERLOGLOG",
    "PARQUET_COMPRESSION",
    "PARTITIONED_BY",
    "PARTITIONED",
    "PARTITIONS",
    "PLUS",
    "PRETTY",
    "PRINCIPALS",
    "PROTECTION",
    "PURGE",
    "QDIGEST",
    "READ",
    "READONLY",
    "REBUILD",
    "RECORDREADER",
    "RECORDWRITER",
    "REGEXP",
    "RELOAD",
    "RELY",
    "RENAME",
    "REPAIR",
    "REPLACE",
    "REPLICATION",
    "RESTRICT",
    "REWRITE",
    "RLIKE",
    "ROLE",
    "ROLES",
    "SCHEMA",
    "SCHEMAS",
    "SECOND",
    "SEMI",
    "SERDE",
    "SERDEPROPERTIES",
    "SERVER",
    "SETS",
    "SHARED",
    "SHOW_DATABASE",
    "SHOW",
    "SKEWED",
    "SNAPSHOT",
    "SORT",
    "SORTED",
    "SSL",
    "STATISTICS",
    "STORED",
    "STREAMTABLE",
    "STRING",
    "STRUCT",
    "TABLES",
    "TBLPROPERTIES",
    "TEMPORARY",
    "TERMINATED",
    "TIMESTAMPTZ",
    "TINYINT",
    "TOUCH",
    "TRANSACTION",
    "TRANSACTIONS",
    "UNARCHIVE",
    "UNDO",
    "UNIONTYPE",
    "UNLOAD",
    "UNLOCK",
    "UNSET",
    "UNSIGNED",
    "URI",
    "USE",
    "UTC",
    "UTCTIMESTAMP",
    "VALIDATE",
    "VALUE_TYPE",
    "VIEW",
    "WHILE",
    "WRITE_COMPRESSION",
    "YEAR",
    "ZONE",
]



"""A list of all SQL key words.

https://docs.microsoft.com/en-us/sql/t-sql/language-elements/reserved-keywords-transact-sql?view=sql-server-ver15
"""

RESERVED_KEYWORDS = [
    "ADD",
    "ALL",
    "ALTER",
    "AND",
    "ANY",
    "APPEND",
    "AS",
    "ASC",
    "AUTHORIZATION",
    "BACKUP",
    "BEGIN",
    "BETWEEN",
    "BREAK",
    "BROWSE",
    "BULK",
    "BY",
    "CASCADE",
    "CASE",
    "CHECK",
    "CHECKPOINT",
    "CLOSE",
    "CLUSTERED",
    "COALESCE",
    "COLLATE",
    "COLUMN",
    "COMMIT",
    "COMPUTE",
    "CONSTRAINT",
    "CONTAINS",
    "CONTAINSTABLE",
    "CONTINUE",
    "CONVERT",
    "CREATE",
    "CROSS",
    "CURRENT_DATE",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "CURRENT_USER",
    "CURRENT",
    "CURSOR",
    "DATABASE",
    "DBCC",
    "DEALLOCATE",
    "DECLARE",
    "DEFAULT",
    "DELETE",
    "DENY",
    "DESC",
    "DISTINCT",
    "DISTRIBUTED",
    "DOUBLE",
    "DROP",
    "DYNAMIC",
    "ELSE",
    "END",
    "ERRLVL",
    "ESCAPE",
    "EXCEPT",
    "EXEC",
    "EXECUTE",
    "EXISTS",
    "EXIT",
    "EXTERNAL",
    "FAST_FORWARD",
    "FETCH",
    "FILE",
    "FILLFACTOR",
    "FOR",
    "FORWARD_ONLY",
    "FOREIGN",
    "FREETEXT",
    "FREETEXTTABLE",
    "FROM",
    "FULL",
    "FULLSCAN",
    "FUNCTION",
    "GLOBAL",
    "GO",
    "GOTO",
    "GRANT",
    "GROUP",
    "HAVING",
    "HOLDLOCK",
    "IDENTITY_INSERT",
    "IDENTITY",
    "IDENTITYCOL",
    "IF",
    "IN",
    "INDEX",
    "INNER",
    "INSERT",
    "INTERSECT",
    "INTO",
    "IS",
    "JOIN",
    "KEY",
    "KEYSET",
    "KILL",
    "LEFT",
    "LIKE",
    "LINENO",
    "LOCAL",
    "MERGE",
    "NATIONAL",
    "NATIVE_COMPILATION",
    "NOCHECK",
    "NONCLUSTERED",
    "NOT",
    "NULL",
    "NULLIF",
    "OF",
    "OFF",
    "OFFSETS",
    "ON",
    "OPEN",
    "OPENDATASOURCE",
    "OPENQUERY",
    "OPENROWSET",
    "OPENXML",
    "OPTIMISTIC",
    "OPTION",
    "OR",
    "ORDER",
    "OUTER",
    "OVER",
    "PERCENT",
    "PIVOT",
    "PLAN",
    "PRIMARY",
    "PRINT",
    "PROC",
    "PROCEDURE",
    "PUBLIC",
    "RAISERROR",
    "READ",
    "READ_ONLY",
    "READTEXT",
    "RECONFIGURE",
    "REFERENCES",
    "REPLICATION",
    "RESAMPLE",
    "RESTORE",
    "RESTRICT",
    "RETURN",
    "REVERT",
    "REVOKE",
    "RIGHT",
    "ROLLBACK",
    "ROWCOUNT",
    "ROWGUIDCOL",
    "RULE",
    "SAVE",
    "SCHEMA",
    "SCROLL",
    "SCROLL_LOCKS",
    "SELECT",
    "SEMANTICKEYPHRASETABLE",
    "SEMANTICSIMILARITYDETAILSTABLE",
    "SEMANTICSIMILARITYTABLE",
    "SESSION_USER",
    "SET",
    "SETUSER",
    "SHUTDOWN",
    "SOME",
    "STATIC",
    "STATISTICS",
    "SYSTEM_USER",
    "TABLE",
    "TABLESAMPLE",
    "TEXTSIZE",
    "THEN",
    "TO",
    "TOP",
    "TRAN",
    "TRANSACTION",
    "TRAN",
    "TRIGGER",
    "TRUNCATE",
    "TRY_CONVERT",
    "TSEQUAL",
    "TYPE_WARNING",
    "UNION",
    "UNIQUE",
    "UNPIVOT",
    "UPDATE",
    "UPDATETEXT",
    "USE",
    "USER",
    "VALUES",
    "VARYING",
    "VIEW",
    "WAITFOR",
    "WHEN",
    "WHERE",
    "WHILE",
    "WITH",
    "WRITETEXT",
]


UNRESERVED_KEYWORDS = [
    "ABORT_AFTER_WAIT",
    "AFTER",
    "ALGORITHM",
    "ALLOW_PAGE_LOCKS",
    "ALLOW_ROW_LOCKS",
    "ALWAYS",
    "ANSI_DEFAULTS",
    "ANSI_NULL_DFLT_OFF",
    "ANSI_NULL_DFLT_ON",
    "ANSI_NULLS",
    "ANSI_PADDING",
    "ANSI_WARNINGS",
    "APPLY",
    "ARITHABORT",
    "ARITHIGNORE",
    "AT",
    "AUTO",
    "BERNOULLI",
    "BLOCKERS",
    "BREAK",
    "CACHE",
    "CALLED",
    "CALLER",
    "CAST",
    "CATCH",
    "CODEPAGE",
    "COLUMN_ENCRYPTION_KEY",
    "COLUMNSTORE",
    "COLUMNSTORE_ARCHIVE",
    "COMMITTED",
    "CONCAT",
    "CONCAT_NULL_YIELDS_NULL",
    "CONTINUE",
    "CONTROL",
    "COMPRESSION_DELAY",
    "CURSOR_CLOSE_ON_COMMIT",
    "CYCLE",
    "DATA_COMPRESSION",
    "DATASOURCE",
    "DATE",
    "DATEFIRST",
    "DATEFORMAT",
    "DEADLOCK_PRIORITY",
    "DELAY",
    "DENSE_RANK",
    "DETERMINISTIC",
    "DISABLE",
    "DISK",  # listed as reserved but functionally unreserved
    "DISTRIBUTION",  # Azure Synapse Analytics specific
    "DROP_EXISTING",
    "DUMP",  # listed as reserved but functionally unreserved
    "ENCRYPTED",
    "ENCRYPTION",
    "ENCRYPTION_TYPE",
    "ERRORFILE",
    "ERRORFILE_DATA_SOURCE",
    "EXPAND",
    "EXPLAIN",  # Azure Synapse Analytics specific
    "EXPLICIT",
    "EXTERNALPUSHDOWN",
    "FAST",
    "FIELDQUOTE",
    "FILESTREAM",
    "FILESTREAM_ON",
    "FILTER",
    "FIPS_FLAGGER",
    "FIRST",
    "FIRSTROW",
    "FMTONLY",
    "FOLLOWING",
    "FOR",
    "FORCE",
    "FORCEPLAN",
    "FORCESCAN",
    "FORCESEEK",
    "FORMAT",
    "FORMATFILE",
    "FORMATFILE_DATA_SOURCE",
    "HASH",
    "HEAP",  # Azure Synapse Analytics specific
    "HIDDEN",
    "HINT",
    "IGNORE",
    "IGNORE_CONSTRAINTS",
    "IGNORE_DUP_KEY",
    "IGNORE_NONCLUSTERED_COLUMNSTORE_INDEX",
    "IGNORE_TRIGGERS",
    "IMPLICIT_TRANSACTIONS",
    "INCLUDE",
    "INCREMENT",
    "INLINE",
    "INSTEAD",
    "INTERVAL",
    "IO",
    "ISOLATION",
    "GENERATED",
    "KEEP",
    "KEEPDEFAULTS",
    "KEEPFIXED",
    "KEEPIDENTITY",
    # LABEL is an Azure Synapse Analytics specific reserved keyword
    # but could break TSQL parsing to add there
    "LABEL",
    "LANGUAGE",
    "LAST",
    "LASTROW",
    "LEVEL",
    "LOAD",  # listed as reserved but functionally unreserved
    "LOCATION",
    "LOCK_TIMEOUT",
    "LOG",
    "LOGIN",
    "LOOP",
    "MAX_DURATION",
    "MAX_GRANT_PERCENT",
    "MASKED",
    "MATCHED",
    "MAX",
    "MAXDOP",
    "MAXERRORS",
    "MAXRECURSION",
    "MAXVALUE",
    "MIN_GRANT_PERCENT",
    "MINUTES",
    "MINVALUE",
    "NEXT",
    "NO",
    "NO_PERFORMANCE_SPOOL",
    "NOCOUNT",
    "NOEXEC",
    "NOEXPAND",
    "NOLOCK",
    "NONE",
    "NOWAIT",
    "NTILE",
    "NUMERIC_ROUNDABORT",
    "OBJECT",
    "OFFSET",
    "ONLINE",
    "OPTIMIZE",
    "OPTIMIZE_FOR_SEQUENTIAL_KEY",
    "OUT",
    "OUTPUT",
    "OWNER",
    "PAD_INDEX",
    "PAGE",
    "PAGLOCK",
    "PARAMETER",
    "PARAMETERIZATION",
    "PARSEONLY",
    "PARTITION",
    "PARTITIONS",
    "PATH",
    "PERCENTILE_CONT",
    "PERCENTILE_DISC",
    "PERSISTED",
    "PRECEDING",
    "PRECISION",  # listed as reserved but functionally unreserved
    "PRIOR",
    "PROFILE",
    "QUERY_GOVERNOR_COST_LIMIT",
    "QUERYTRACEON",
    "QUOTED_IDENTIFIER",
    "RANDOMIZED",
    "RANGE",
    "RANK",
    "RAW",
    "READCOMMITTED",
    "READCOMMITTEDLOCK",
    "READONLY",
    "READPAST",
    "READUNCOMMITTED",
    "RECEIVE",
    "RECOMPILE",
    "RECURSIVE",
    "REMOTE_PROC_TRANSACTIONS",
    "RENAME",  # Azure Synapse Analytics specific
    "REPEATABLE",
    "REPEATABLEREAD",
    "REPLACE",
    "REPLICATE",  # Azure Synapse Analytics
    "RESPECT",
    "RESULT_SET_CACHING",  # Azure Synapse Analytics specific
    "RESUMABLE",
    "RETURNS",
    "ROBUST",
    "ROLE",
    "ROUND_ROBIN",  # Azure Synapse Analytics specific
    "ROW",
    "ROW_NUMBER",
    "ROWGUIDCOL",
    "ROWLOCK",
    "ROWS",
    "S",
    "SCALEOUTEXECUTION",
    "SCHEMABINDING",
    "SECURITYAUDIT",  # listed as reserved but functionally unreserved
    "SELF",
    "SETERROR",
    "SEQUENCE",
    "SEQUENCE_NUMBER",
    "SERIALIZABLE",
    "SERVER",
    "SHOWPLAN_ALL",
    "SHOWPLAN_TEXT",
    "SHOWPLAN_XML",
    "SINGLE_BLOB",
    "SINGLE_CLOB",
    "SINGLE_NCLOB",
    "SNAPSHOT",
    "SORT_IN_TEMPDB",
    "SOURCE",
    "SPARSE",
    "SPATIAL_WINDOW_MAX_CELLS",
    "START",
    "STATISTICS_INCREMENTAL",
    "STATISTICS_NORECOMPUTE",
    "STRING_AGG",
    "SYNONYM",
    "SWITCH",
    "SYSTEM",
    "TABLOCK",
    "TABLOCKX",
    "TAKE",
    "TARGET",
    "TEXTIMAGE_ON",
    "THROW",
    "TIES",
    "TIME",
    "TIMEOUT",
    "TIMESTAMP",
    "TRANSACTION_ID",
    "TRUNCATE_TARGET",  # Azure Synapse Analytics specific
    "TRY",
    "TYPE",
    "UPDLOCK",
    "UNBOUNDED",
    "UNCOMMITTED",
    "UNKNOWN",
    "USER_DB",  # Azure Synapse Analytics specific, deprecated
    "USING",
    "VALUE",
    "VIEW_METADATA",
    "WAITFOR",
    "WAIT_AT_LOW_PRIORITY",
    "WITHIN",
    "WHILE",
    "WORK",
    "XACT_ABORT",
    "XLOCK",
    "XML",
    "ZONE",
]



"""Defines the LintedDir class.

This stores the idea of a collection of linted files at a single start path

"""

from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
    overload,
)
from typing_extensions import Literal


from sqlfluff.core.errors import (
    CheckTuple,
)
from sqlfluff.core.parser.segments.base import BaseSegment

from sqlfluff.core.linter.linted_file import LintedFile


class LintedDir:
    """A class to store the idea of a collection of linted files at a single start path.

    A LintedDir may contain files in subdirectories, but they all share
    a common root.
    """

    def __init__(self, path: str) -> None:
        self.files: List[LintedFile] = []
        self.path: str = path

    def add(self, file: LintedFile) -> None:
        """Add a file to this path."""
        self.files.append(file)

    @overload
    def check_tuples(
        self, by_path: Literal[False]
    ) -> List[CheckTuple]:  # pragma: no cover
        """Return a List of CheckTuples when by_path is False."""
        ...

    @overload
    def check_tuples(
        self, by_path: Literal[True]
    ) -> Dict[str, List[CheckTuple]]:  # pragma: no cover
        """Return a Dict of paths and CheckTuples when by_path is True."""
        ...

    @overload
    def check_tuples(self, by_path: bool = False):  # pragma: no cover
        """Default overload method."""
        ...

    def check_tuples(
        self, by_path=False, raise_on_non_linting_violations=True
    ) -> Union[List[CheckTuple], Dict[str, List[CheckTuple]]]:
        """Compress all the tuples into one list.

        NB: This is a little crude, as you can't tell which
        file the violations are from. Good for testing though.
        For more control set the `by_path` argument to true.
        """
        if by_path:
            return {
                file.path: file.check_tuples(
                    raise_on_non_linting_violations=raise_on_non_linting_violations
                )
                for file in self.files
            }
        else:
            tuple_buffer: List[CheckTuple] = []
            for file in self.files:
                tuple_buffer += file.check_tuples(
                    raise_on_non_linting_violations=raise_on_non_linting_violations
                )
            return tuple_buffer

    def num_violations(self, **kwargs) -> int:
        """Count the number of violations in the path."""
        return sum(file.num_violations(**kwargs) for file in self.files)

    def get_violations(self, **kwargs) -> list:
        """Return a list of violations in the path."""
        buff: list = []
        for file in self.files:
            buff += file.get_violations(**kwargs)
        return buff

    def violation_dict(self, **kwargs) -> Dict[str, list]:
        """Return a dict of violations by file path."""
        return {file.path: file.get_violations(**kwargs) for file in self.files}

    def stats(self) -> Dict[str, int]:
        """Return a dict containing linting stats about this path."""
        return dict(
            files=len(self.files),
            clean=sum(file.is_clean() for file in self.files),
            unclean=sum(not file.is_clean() for file in self.files),
            violations=sum(file.num_violations() for file in self.files),
        )

    def persist_changes(
        self, formatter: Any = None, fixed_file_suffix: str = "", **kwargs
    ) -> Dict[str, Union[bool, str]]:
        """Persist changes to files in the given path.

        This also logs the output as we go using the formatter if present.
        """
        # Run all the fixes for all the files and return a dict
        buffer: Dict[str, Union[bool, str]] = {}
        for file in self.files:
            if file.num_violations(fixable=True, **kwargs) > 0:
                buffer[file.path] = file.persist_tree(suffix=fixed_file_suffix)
                result = buffer[file.path]
            else:  # pragma: no cover TODO?
                buffer[file.path] = True
                result = "SKIP"

            if formatter:
                formatter.dispatch_persist_filename(filename=file.path, result=result)
        return buffer

    @property
    def tree(self) -> Optional[BaseSegment]:
        """A convenience method for when there is only one file and we want the tree."""
        if len(self.files) > 1:  # pragma: no cover
            raise ValueError(
                ".tree() cannot be called when a LintedDir contains more than one file."
            )
        return self.files[0].tree
